#include "sokodan.h"
#include "PositionX.h"
#include "Position.h"
#include <string.h>

#undef DEBUG

void Position_init (void)
{
  Position_Pool = g_mem_chunk_create (Position, POSITION_PREALLOC, G_ALLOC_AND_FREE);
}

PositionPtr Position_new (void)
{
  return g_chunk_new (Position, Position_Pool);
}

PositionPtr Position_duplicate (const PositionPtr pos)
{
  PositionPtr duplicata = Position_new ();
  memcpy (duplicata, pos, sizeof (Position));
  return duplicata;
}

void gPosition_delete (gpointer data, gpointer extra)
{
#ifdef DEBUG
  sprintf ((char *) data, "GARBAGE!!!0123456789"); /* hum... */
#endif
  g_chunk_free ((PositionPtr) data, Position_Pool);
  g_assert (extra == NULL);
}


/*
  +----------------------------------------------------------------------------
  | NAME:     Position_do_move
  | PURPOSE:  Computes the resulting position after a move.
  | INPUT:    A pointer to the initial position, a pointer to the move, and a
  |           pointer to the resulting position.
  | OUTPUT:   None.
  | NOTES:    None.
  +----------------------------------------------------------------------------
*/
void Position_do_move (PositionPtr init,
                       MovePtr     move,
                       PositionPtr final)
{
  Coordinate x, y;
  gint8      dx, dy;

  /* In the begining, final was like init */
  memcpy (final, init, sizeof (Position));

  /* Do the move */
  x  = move->x;
  y  = move->y;
  dx = Delta[ move->dir ].dx;
  dy = Delta[ move->dir ].dy;
  REMOVE (GEM, final->grid[ x + dx ][ y + dy ]);
  ADD    (GEM, final->grid[ x + 2*dx ][ y + 2*dy ]);
  final->player_x = x + dx;
  final->player_y = y + dy;
  _Position_compute_hash_value (final);
  final->previous = init;

  if (!CONTAIN (ACCESSIBLE, final->grid[ x + 2*dx ][ y + 2*dy ]))
    /* If the case where the gem is now was not marked as accessible,
     * simply mark the case where it was. */
  {
    _rec_mark_access (final, x + dx, y + dy);
  }
  else /* general case */
  {
    /* Remove any ACCESSIBLE attribute */
    _Position_unmark_access (final);
      
    /* Recompute accessible cases */
    _Position_mark_access (final);
  }
}

/*
  +----------------------------------------------------------------------------
  | NAME:     gPosition_hash
  | PURPOSE:  Hashes a position.
  | INPUT:    A pointer (to a position).
  | OUTPUT:   A unsigned integer (the hash value).
  | NOTES:    Nothing is done here, since the hash value is assumed to have
  |           already been computed.
  +----------------------------------------------------------------------------
*/
guint gPosition_hash (gconstpointer _pos)
{
  return ((PositionPtr) _pos)->hash_value;
}

/*
  +----------------------------------------------------------------------------
  | NAME:     _Position_compute_hash_value
  | PURPOSE:  Actually computes the hash value of a position.
  | INPUT:    A pointer to a position.
  | OUTPUT:   Nothing.
  | NOTES:    The hash value is stored in the position itself.
  +----------------------------------------------------------------------------
*/
void _Position_compute_hash_value (PositionPtr pos)
{
  guint *gems;
  gint i;
  Coordinate x, y;

  gems = (guint *) g_malloc (Number_of_targets * sizeof (guint));

  i = 0;
  for (x = 0; x < WIDTH; x++)
    for (y = 0; y < HEIGHT; y++)
      if (CONTAIN (GEM, pos->grid[ x ][ y ]))
        gems[ i++ ] = y + x * HEIGHT;

  g_assert (i == Number_of_targets);

  for (i = Number_of_targets-1; i>0; i--)
    gems[ i ] -= gems[ i-1 ];

  pos->hash_value = _Position_hash (gems);

  g_free (gems);
}

/*
  +----------------------------------------------------------------------------
  | NAME:     _Position_hash
  | PURPOSE:  Computes the hash value of a position (low-level function).
  | INPUT:    An array of positive integers describing where the gems are.
  | OUTPUT:   The hash value (as an unsigned integer).
  | NOTES:    
  +----------------------------------------------------------------------------
*/
static guint _Position_hash (guint gems[])
{
  gint result = 0;
  gint i, t;
  
  for (i = 0; i < Number_of_targets; i++)
  {
    result ^= gems[ i ];
    t = result >> (8 * sizeof (gint) - _HASH_OFFSET);
    result = (result << _HASH_OFFSET) | t;
  }

  return result;
}

/*
  +----------------------------------------------------------------------------
  | NAME:     gPosition_compare
  | PURPOSE:  Says whether two positions are equal or not.
  | INPUT:    Two pointers (to two positions).
  | OUTPUT:   An integer (describing a boolean).
  | NOTES:    The gconstpointers are expected to be PositionPtr, and the return
  |           value is TRUE if the position are equivalent.
  +----------------------------------------------------------------------------
*/
gint gPosition_compare (gconstpointer _pos_a, gconstpointer _pos_b)
{
  PositionPtr pos_a = (PositionPtr) _pos_a;
  PositionPtr pos_b = (PositionPtr) _pos_b;

  if (memcmp (&pos_a->grid[ 0 ][ 0 ], 
              &pos_b->grid[ 0 ][ 0 ], 
              HEIGHT * WIDTH * sizeof (Case)))
    return FALSE;

  return TRUE;
}

/*
  +----------------------------------------------------------------------------
  | NAME:     _Position_mark_access, _Position_unmark_access and
  |           _rec_mark_access
  | PURPOSE:  Marks as ACCESSIBLE all the cases in the grid where the player
  |           can go, or clear the ACCESSIBLE flag.
  | INPUT:    A pointer to a position.
  | OUTPUT:   None.
  | NOTES:    
  +----------------------------------------------------------------------------
*/
static void _Position_mark_access (PositionPtr pos)
{
  g_assert (pos != NULL);

  _rec_mark_access (pos, pos->player_x, pos->player_y);
}

static void _Position_unmark_access (PositionPtr pos)
{
  Coordinate x, y;
  g_assert (pos != NULL);

  /* Remove any ACCESSIBLE attribute */
  for (x = 0; x < WIDTH; x++)
    for (y = 0; y < HEIGHT; y++)
      /* Is it better with this test? */
      if (CONTAIN (ACCESSIBLE, pos->grid[ x ][ y ]))
        REMOVE (ACCESSIBLE, pos->grid[ x ][ y ]);
}

static void _rec_mark_access (PositionPtr pos,
                              Coordinate  x,
                              Coordinate  y)
{
  if ( CONTAIN (ACCESSIBLE, pos->grid[ x ][ y ]) )
    return;
  
  ADD (ACCESSIBLE, pos->grid[ x ][ y ]);

  /* Try to go right */
  if (x+1 < WIDTH
      && CAN_GO (pos->grid[ x+1 ][ y ]) )
  {
    _rec_mark_access (pos, x+1, y);
  }

  /* Try to go left */
  if (x >= 1 /* Beware : x is unsigned, so x-1 might be unsigned too... */
      && CAN_GO (pos->grid[ x-1 ][ y ]) )
  {
    _rec_mark_access (pos, x-1, y);
  }

  /* Try to go down */
  if (y+1 < HEIGHT
      && CAN_GO (pos->grid[ x ][ y+1 ]) )
  {
    _rec_mark_access (pos, x, y+1);
  }

  /* Try to go up */
  if (y >= 1
      && CAN_GO (pos->grid[ x ][ y-1 ]) )
  {
    _rec_mark_access (pos, x, y-1);
  }
}

/*
  +----------------------------------------------------------------------------
  | NAME:     Position_is_winning
  | PURPOSE:  Says whether the position is winning or not.
  | INPUT:    A pointer to a position.
  | OUTPUT:   A boolean.
  | NOTES:    First we used Winning_hash. Next, we should compare (memcmp) to
  |           a Winning_position, rather than test each case.
  +----------------------------------------------------------------------------
*/
gboolean Position_is_winning (const PositionPtr pos)
{
  Coordinate x, y;

  if (pos->hash_value != Winning_hash)
    return FALSE;

  for (x = 0; x < WIDTH; x++)
    for (y = 0; y < HEIGHT; y++)
      if (CONTAIN (TARGET, pos->grid[ x ][ y ])
          && !CONTAIN (GEM, pos->grid[ x ][ y ]))
        return FALSE;

  return TRUE;
}

/*
  +----------------------------------------------------------------------------
  | NAME:     solve_init
  | PURPOSE:  Performs some important initializations before solving the pb.
  |           Returns a pointer to a copy of Problem, with "decorations".
  | INPUT:    None (cf Problem).
  | OUTPUT:   None (cf Number_of_targets and Winning_hash).
  | NOTES:    Number_of_targets and Winning_hash are computed. Problem is
  |           prepared for its insertion into the big hash table.
  +----------------------------------------------------------------------------
*/
void solve_init (void)
{
  guint *targets;
  gint i;
  Coordinate x, y;

  /* Compute Number_of_targets before any checksum calculation */
  Number_of_targets = 0;
    
  for (x = 0; x < WIDTH; x++)
    for (y = 0; y < HEIGHT; y++)
      if (CONTAIN (TARGET, Problem.grid[ x ][ y ]))
        Number_of_targets++;
  
  /* Compute the Winning_hash */
  targets = (guint *) g_malloc (Number_of_targets * sizeof (guint));

  i = 0;
  for (x = 0; x < WIDTH; x++)
    for (y = 0; y < HEIGHT; y++)
      if (CONTAIN (TARGET, Problem.grid[ x ][ y ]))
        targets[ i++ ] = y + x * HEIGHT;

  g_assert (i == Number_of_targets);

  for (i = Number_of_targets-1; i>0; i--)
    targets[ i ] -= targets[ i-1 ];

  Winning_hash = _Position_hash (targets);

  g_free (targets);

  /* Add some important stuff to Problem. */
  _Position_mark_access (&Problem);
  _Position_compute_hash_value (&Problem);
  Problem.previous = NULL;
}


/*
  +----------------------------------------------------------------------------
  | NAME:     solve_term
  | PURPOSE:  Restores some stuff before returning to edit mode.
  | INPUT:    None.
  | OUTPUT:   None.
  | NOTES:    
  +----------------------------------------------------------------------------
*/
void solve_term (void)
{
  /* Clear Problem */
  _Position_unmark_access (&Problem);
}
